// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by
//            pkgmaker 0.1
//            Thursday, December 11 1997
//
// Module: Statistic
// Author: Hassan Siddiqui
//
// Description: Computes simple statistics on-the-fly from a given input data list.
//
// Comments:
//
// Modifications: See ChangeLog
//
// $Id: Statistic.cc,v 1.7 2009/09/01 15:50:54 aibarra Exp $

/* General */
#include <string> // put this up at the top as order matters
#include <iostream>
#include <cmath>
#include <list>
#include<cassert>

/* SAS-specific */
#include "StatisticElement.h"

/* Local */
#include "Statistic.h"
#include "errstr.h"
#include "Tracer.h"

/* Macro for computing stats for various data types */
#define COMPUTE_STATISTIC(TYPE)\
TYPE* data = col->data()->TYPE ## Data();\
_min = _max = StatisticElement((*rowranges.begin()).startRow, data[(*rowranges.begin()).startRow]);\
for (vector<RowRange>::const_iterator index = rowranges.begin();\
     index != rowranges.end(); ++index)\
{\
  for (unsigned long row = (*index).startRow; row <= (*index).stopRow; ++row)\
  {\
    _size++;\
    updateBasicStatistics(StatisticElement(row, data[row]));\
  }\
}\
_meanset = true;\
for (vector<RowRange>::const_iterator index = rowranges.begin();\
     index != rowranges.end(); ++index)\
{\
   for (unsigned long row = (*index).startRow; row <= (*index).stopRow; ++row)\
     updateDeviation(StatisticElement(row, data[row]));\
}\


/* Constructors */
Statistic::Statistic():
  _size(0),
  _sum(0),
  _sumOfSquares(0),
  _deviation(0),
  _min(0,0),
  _max(0,0),
  _meanset(false)
{PROC; ENDPROC};

Statistic::Statistic(const Table& table,
		      const string& name,
		      const vector<RowRange>& rowranges):
  _size(0),
  _sum(0),
  _sumOfSquares(0),
  _deviation(0),
  _min(0,0),
  _max(0,0),
  _meanset(false)
{

  const Column* col = table.column(name);

  _name = col->qualifiedName();

  // Compute statistics column based on column type,
  // in absence of automatic type conversions provided by the DAL.

  GENERIC_OPERATION(COMPUTE_STATISTIC, (int) col->dataType());

};

Statistic::Statistic(const Column * col):
  _size(0),
  _sum(0),
  _sumOfSquares(0),
  _deviation(0),
  _min(0,0),
  _max(0,0),
  _meanset(false)
{
  _name = col->qualifiedName();

  /* Compute dummy row range*/
  RowRange rowrange;
  rowrange.startRow = 0;
  rowrange.stopRow  = col->rows()-1;

  vector<RowRange> rowranges;

  rowranges.push_back(rowrange);

  /* Compute statistics column based on column type,
     in absence of automatic type conversions provided by the DAL. */

  GENERIC_OPERATION(COMPUTE_STATISTIC, (int) col->dataType());

};


void Statistic::updateBasicStatistics(const StatisticElement& element)
{
  if (element < _min)
    {
      _min = element;
    }

  if (element > _max)
    {
      _max = element;
    }

  double value = element.value();

  // Update Mean
  _sum += value;

  // Update RMS
  _sumOfSquares += value * value;

}

void Statistic::updateDeviation(const StatisticElement& element)
{
  assert(_meanset); // Ensure that there is a mean value computed - This is a bit messy, fix later

  // Don't do anything if there are no data.
  if (_size == 0) return;

  double diff = element.value() - (_sum/_size);
  _deviation += diff * diff;
}


/* Accessors */

const StatisticElement& Statistic::max()  const   {PROC; RETURN _max;}
const StatisticElement& Statistic::min() const    {PROC; RETURN _min;}

/* AI gcc4 porting
  * error: type qualifiers ignored on function return type
  * => remove const
  */
 double Statistic::mean() const              {PROC; assert(_size > 0); RETURN (_sum/_size);}
 double Statistic::sum() const               {PROC; RETURN _sum;}
 double Statistic::rootMeanSquared()   const {PROC; assert(_size > 0); RETURN sqrt(_sumOfSquares/_size);}
 double Statistic::standardDeviation() const {PROC; assert(_size > 0); RETURN sqrt(_deviation/_size);}
 unsigned long Statistic::size() const       {PROC; RETURN _size;}


/* Output stream method */
void Statistic::write(ostream& stream) const
{
  PROC;

  if (_size > 0)
    {
      stream
	<< "Statistics for column " << _name << " : \n"
	<< " MIN "  << _min
	<< " MAX "  << _max
	<< " MEAN " << mean() << endl
	<< " RMS "  << rootMeanSquared() << endl
	<< " STDDEV " << standardDeviation() << endl
	<< " SUM "  << _sum  << endl
	<< " SIZE " << _size
	<< "\n"
	<< ends;

      ALTRETURN;
    }

  stream << "NULL" << ends;

  ENDPROC;
}

/* I/O Stream functions */
ostream & operator << (ostream & stream, const Statistic & statistic)
{
  PROC;
  statistic.write(stream);
  RETURN stream;
}


